
from nflows.transforms.permutations import RandomPermutation
import torch

from nessai.flows.utils import reset_permutations


def test_weight_reset_permutation():
    """Test to make sure random permutation is reset correctly"""
    x = torch.arange(10).reshape(1, -1)
    m = RandomPermutation(features=10)
    y_init, _ = m(x)
    p = m._permutation.numpy()
    m.apply(reset_permutations)
    y_reset, _ = m(x)
    assert not (p == m._permutation.numpy()).all()
    assert not (y_init.numpy() == y_reset.numpy()).all()



# -*- coding: utf-8 -*-
"""
Proposal methods for used in the nested sampler.
"""
from .analytic import AnalyticProposal
from .flowproposal import FlowProposal
from .rejection import RejectionProposal

__all__ = ["AnalyticProposal",
           "FlowProposal",
           "RejectionProposal"
           ]



import numpy as np
import pytest

from nessai.model import Model


@pytest.fixture(scope='function')
def empty_model():
    return Model()


def test_dims_no_names(empty_model):
    """Test the behaviour dims when names is empty"""
    assert empty_model.dims is None


def test_dims(empty_model):
    """Ensure dims are correct"""
    empty_model.names = ['x', 'y']
    assert empty_model.dims == 2


def test_lower_bounds(empty_model):
    """Check the lower bounds are correctly set"""
    empty_model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    assert (empty_model.lower_bounds == [-1, -1]).all()


def test_upper_bounds(empty_model):
    """Check the upper bounds are correctly set"""
    empty_model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    assert (empty_model.upper_bounds == [1, 1]).all()


def test_new_point(model):
    """
    Test the default method for generating a new point with the bounds.

    Uses the model defined in `conftest.py` with bounds [-5, 5] for
    x and y.
    """
    new_point = model.new_point()
    log_q = model.new_point_log_prob(new_point)
    assert (new_point['x'] < 5) & (new_point['y'] > -5)
    assert (new_point['y'] < 5) & (new_point['y'] > -5)
    assert log_q == 0


def test_new_point_multiple(model):
    """
    Test drawing multiple new points from the model

    Uses the model defined in `conftest.py` with bounds [-5, 5] for
    x and y.
    """
    new_points = model.new_point(N=100)
    log_q = model.new_point_log_prob(new_points)
    assert new_points.size == 100
    assert all(np.isfinite(new_points['logP']))
    assert all(new_points['x'] < 5) & all(new_points['x'] > -5)
    assert all(new_points['y'] < 5) & all(new_points['y'] > -5)
    assert (log_q == 0).all()


def test_likelihood_evaluations(model):
    """
    Test `evaluate_log_likelihood` and ensure the counter increases.
    """
    new_points = model.new_point(N=1)
    log_l = model.evaluate_log_likelihood(new_points)

    assert log_l.size == 1
    assert model.likelihood_evaluations == 1


def test_verify_new_point():
    """
    Test `Model.verify_model` and ensure a model with an ill-defined
    prior function raises the correct error
    """
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return -np.inf

    model = TestModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert 'valid point' in str(excinfo.value)


def test_verify_log_prior():
    """
    Test `Model.verify_model` and ensure a model without a log-prior
    function raises the correct error
    """
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

    model = TestModel()

    with pytest.raises(RuntimeError):
        model.verify_model()


@pytest.mark.parametrize("log_p", [np.inf, -np.inf])
def test_verify_log_prior_finite(log_p):
    """
    Test `Model.verify_model` and ensure a model with a log-prior that
    only returns inf function raises the correct error
    """
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return log_p

    model = TestModel()

    with pytest.raises(RuntimeError):
        model.verify_model()


def test_verify_log_likelihood():
    """
    Test `Model.verify_model` and ensure a model without a log-likelihood
    function raises the correct error
    """
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return 0.

    model = TestModel()

    with pytest.raises(RuntimeError):
        model.verify_model()


def test_verify_no_names():
    """
    Test `Model.verify_model` and ensure a model without names
    function raises the correct error
    """
    class TestModel(Model):

        def __init__(self):
            self.names = []
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}

    model = TestModel()

    with pytest.raises(ValueError):
        model.verify_model()


def test_verify_no_bounds():
    """
    Test `Model.verify_model` and ensure a model without bounds
    function raises the correct error
    """
    class TestModel(Model):

        def __init__(self):
            self.names = ['x', 'y']
            self.bounds = {}

    model = TestModel()

    with pytest.raises(ValueError) as excinfo:
        model.verify_model()

    assert 'Bounds are not set' in str(excinfo.value)


def test_verify_incomplete_bounds():
    """
    Test `Model.verify_model` and ensure a model without bounds
    function raises the correct error
    """
    class TestModel(Model):

        def __init__(self):
            self.names = ['x', 'y']
            self.bounds = {'x': [-5, 5]}

    model = TestModel()

    with pytest.raises(RuntimeError):
        model.verify_model()



# -*- coding: utf-8 -*-
"""
Object and functions to handle training the normalising flow.
"""
import copy
import json
import logging
import numpy as np
import os
import shutil
import torch
from torch.nn.utils import clip_grad_norm_

from .flows import setup_model, reset_weights, reset_permutations
from .plot import plot_loss
from .utils import FPJSONEncoder, compute_minimum_distances

logger = logging.getLogger(__name__)


def update_config(d):
    """
    Update the configuration dictionary to include the defaults.

    Notes
    -----
    The default configuration is::

        lr=0.001
        annealing=False
        batch_size=100
        val_size=0.1
        max_epochs=500
        patience=20
        noise_scale=0.0
        model_config=default_model

    where ``default model`` is::

        n_neurons=32
        n_blocks=4
        n_layers=2
        ftype='RealNVP'
        device_tag='cpu'
        kwargs={batch_norm_between_layers=True, linear_transform='lu'}

    The kwargs can contain any additional keyword arguments that are specific
    to the type of flow being used.

    Parameters
    ----------
    d : dict
        Dictionary with configuration

    Returns
    -------
    dict
        Dictionary with updated default configuration
    """
    default_model = dict(n_inputs=None, n_neurons=32, n_blocks=4, n_layers=2,
                         ftype='RealNVP', device_tag='cpu', flow=None,
                         kwargs=dict(batch_norm_between_layers=True,
                                     linear_transform='lu'))

    default = dict(lr=0.001,
                   annealing=False,
                   clip_grad_norm=5,
                   batch_size=100,
                   val_size=0.1,
                   max_epochs=500,
                   patience=20,
                   noise_scale=0.0)

    if d is None:
        default['model_config'] = default_model
    else:
        if not isinstance(d, dict):
            raise TypeError('Must pass a dictionary to update the default '
                            'trainer settings')
        else:
            default.update(d)
            default_model.update(d.get('model_config', {}))
            default['model_config'] = default_model

    return default


class FlowModel:
    """
    Object that contains the normalsing flows and handles training and data
    pre-processing.

    Does NOT use stuctured arrays for live points, \
            :obj:`~nessai.proposal.base.Proposal`
    object should act as the interface between structured used by the sampler
    and unstructured arrays of live points used for training.

    Parameters
    ----------
    config : dict, optional
        Configuration used for the normalising flow. If None, default values
        are used.
    output : str, optional
        Path for output, this includes weights files and the loss plot.
    """
    def __init__(self, config=None, output='./'):
        self.model = None
        self.initialised = False
        self.output = output
        self.setup_from_input_dict(config)
        self.weights_file = None

    def save_input(self, config, output_file=None):
        """
        Save the dictionary used as an inputs as a JSON file in the output
        directory.

        Parameters
        ----------
        config : dict
            Dictionary to save.
        ouput_file : str, optional
            File to save the config to.
        """
        config = copy.deepcopy(config)
        if output_file is None:
            output_file = self.output + "flow_config.json"
        for k, v in list(config.items()):
            if type(v) == np.ndarray:
                config[k] = np.array_str(config[k])
        for k, v in list(config['model_config'].items()):
            if type(v) == np.ndarray:
                config['model_config'][k] = \
                    np.array_str(config['model_config'][k])

        if 'flow' in config['model_config']:
            config['model_config']['flow'] = \
                str(config['model_config']['flow'])

        with open(output_file, "w") as f:
            json.dump(config, f, indent=4, cls=FPJSONEncoder)

    def setup_from_input_dict(self, config):
        """
        Setup the trainer from a dictionary, all keys in the dictionary are
        added as methods to the ocject. Input is automatically saved.

        Parameters
        ----------
        config : dict
            Dictionary with parameters that are used to update the defaults.
        """
        config = update_config(config)
        logger.debug(f'Flow configuration: {config}')
        for key, value in config.items():
            setattr(self, key, value)
        self.save_input(config)

    def update_mask(self):
        """Method to update the ask upon calling ``initialise``

        By default the mask is left unchanged.
        """
        pass

    def get_optimiser(self):
        """
        Get the optimiser and ensure it is always correctly intialised.

        Returns
        -------
        :obj:`torch.optim.Adam`
            Instance of the Adam optimiser from torch.optim
        """
        if self.model is None:
            raise RuntimeError('Cannot initialise optimiser before model')
        return torch.optim.Adam(self.model.parameters(),
                                lr=self.lr, weight_decay=1e-6)

    def initialise(self):
        """
        Initialise the model and optimiser.

        This includes:

            - Updating the model configuration
            - Initialising the normalising flow
            - Initialiseing the optimiser
        """
        self.update_mask()
        self.model, self.device = setup_model(self.model_config)
        self.optimiser = self.get_optimiser()
        self.initialised = True

    def prep_data(self, samples, val_size, batch_size):
        """
        Prep data and return dataloaders for training

        Parameters
        ----------
        samples : array_like
            Array of samples to split in to training and validation.
        val_size : float
            Float between 0 and 1 that defines the fraction of data used for
            validation.
        batch_size : int
            Batch size used when contructing dataloaders.

        Returns
        -------
        train_loader, val_loader : :obj:`torch.utils.data.Dataloader`
            Dataloaders with training and validaiton data
        """
        idx = np.random.permutation(samples.shape[0])
        samples = samples[idx]

        logger.debug("N input samples: {}".format(len(samples)))

        # setup data loading
        n = int((1 - val_size) * samples.shape[0])
        x_train, x_val = samples[:n], samples[n:]
        logger.debug(f'{x_train.shape} training samples')
        logger.debug(f'{x_val.shape} validation samples')

        if not type(batch_size) is int:
            if batch_size == 'all':
                batch_size = x_train.shape[0]
            else:
                raise RuntimeError(f'Unknown batch size: {batch_size}')
        train_tensor = torch.from_numpy(x_train.astype(np.float32))
        train_dataset = torch.utils.data.TensorDataset(train_tensor)
        train_loader = torch.utils.data.DataLoader(train_dataset,
                                                   batch_size=batch_size,
                                                   shuffle=True)

        val_tensor = torch.from_numpy(x_val.astype(np.float32))
        val_dataset = torch.utils.data.TensorDataset(val_tensor)
        val_loader = torch.utils.data.DataLoader(val_dataset,
                                                 batch_size=x_val.shape[0],
                                                 shuffle=False)

        return train_loader, val_loader

    def _train(self, loader, noise_scale=0.0):
        """
        Loop over the data and update the weights

        Parameters
        ----------
        loader : :obj:`torch.util.data.Dataloader`
            Dataloader with data to train on
        noise_scale : float, optional
            Scale of Gaussian noise added to data.

        Returns
        -------
        float
            Mean of training loss for each batch.
        """
        model = self.model
        model.train()
        train_loss = 0

        if hasattr(model, 'loss_function'):
            loss_fn = model.loss_function
        else:
            def loss_fn(data):
                return -model.log_prob(data).mean()

        for idx, data in enumerate(loader):
            if noise_scale:
                data[0] += noise_scale * torch.randn_like(data[0])
            data = data[0].to(self.device)
            self.optimiser.zero_grad()
            loss = loss_fn(data)
            train_loss += loss.item()
            loss.backward()
            if self.clip_grad_norm:
                clip_grad_norm_(model.parameters(), self.clip_grad_norm)
            self.optimiser.step()

        if self.annealing:
            self.scheduler.step()

        return train_loss / len(loader)

    def _validate(self, loader):
        """
        Loop over the data and get validation loss

        Parameters
        ----------
        loader : :obj:`torch.util.data.Dataloader`
            Dataloader with data to validate on

        Returns
        -------
        float
            Mean of training loss for each batch.
        """
        model = self.model
        model.eval()
        val_loss = 0

        if hasattr(model, 'loss_function'):
            loss_fn = model.loss_function
        else:
            def loss_fn(data):
                return -model.log_prob(data).mean()

        for idx, data in enumerate(loader):
            data = data[0].to(self.device)
            with torch.no_grad():
                val_loss += loss_fn(data).item()

        return val_loss / len(loader)

    def train(self, samples, max_epochs=None, patience=None, output=None,
              val_size=None, plot=True):
        """
        Train the flow on a set of samples.

        Allows for training parameters to specified instead of those
        given in initial config.

        Parameters
        ----------
        samples : ndarray
            Unstructured numpy array containing data to train on
        max_epochs : int, optional
            Maxinum number of epochs that is used instead of value
            in the configuration.
        patience : int, optional
            Patience in number of epochs that is used instead of value
            in the configuration.
        val_size : float, optional
           Fraction of the samples to use for validation
        output : str, optional
            Path to output directory that is used instead of the path
            specified when the object is initialised
        plot : bool, optional
            Boolean to enable or disable plotting the loss function
        """
        if not self.initialised:
            logger.info("Initialising")
            self.initialise()

        if output is None:
            output = self.output
        else:
            os.makedirs(output, exist_ok=True)

        if val_size is None:
            val_size = self.val_size

        if self.noise_scale == 'adaptive':
            noise_scale = 0.1 * np.std(compute_minimum_distances(samples))
            logger.debug(f'Using adaptive scale: {noise_scale:.3f}')
        else:
            noise_scale = self.noise_scale

        train_loader, val_loader = self.prep_data(samples, val_size=val_size,
                                                  batch_size=self.batch_size)

        if max_epochs is None:
            max_epochs = self.max_epochs
        if self.annealing:
            self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                    self.optimiser, max_epochs)
        if patience is None:
            patience = self.patience
        best_epoch = 0
        best_val_loss = np.inf
        best_model = copy.deepcopy(self.model.state_dict())
        logger.info("Starting training")
        logger.info("Training parameters:")
        logger.info(f"Max. epochs: {max_epochs}")
        logger.info(f"Patience: {patience}")

        if plot:
            history = dict(loss=[], val_loss=[])

        current_weights_file = output + 'model.pt'
        logger.debug(f'Training with {samples.shape[0]} samples')
        for epoch in range(1, max_epochs + 1):

            loss = self._train(train_loader, noise_scale=noise_scale)
            val_loss = self._validate(val_loader)
            if plot:
                history['loss'].append(loss)
                history['val_loss'].append(val_loss)

            if val_loss < best_val_loss:
                best_epoch = epoch
                best_val_loss = val_loss
                best_model = copy.deepcopy(self.model.state_dict())

            if not epoch % 50:
                logger.info(
                    f'Epoch {epoch}: loss: {loss:.3} val loss: {val_loss:.3}')

            if epoch - best_epoch > patience:
                logger.info(f'Epoch {epoch}: Reached patience')
                break

        self.model.load_state_dict(best_model)
        self.save_weights(current_weights_file)
        self.model.eval()

        if plot:
            plot_loss(epoch, history, filename=output + '/loss.png')

    def save_weights(self, weights_file):
        """
        Save the weights file. If the file already exists move it to
        ``<weights_file>.old`` and then save the file.

        Parameters
        ----------
        weights_file : str
            Path to to file to save weights. Recommended file type is ``.pt``.
        """
        if os.path.exists(weights_file):
            shutil.move(weights_file, weights_file + '.old')

        torch.save(self.model.state_dict(), weights_file)
        self.weights_file = weights_file

    def load_weights(self, weights_file):
        """
        Load weights for the model and initialiases the model if it is not
        intialised. The weights_file attribute is also updated.

        Model is loaded in evaluation mode (``model.eval()``)

        Parameters
        ----------
        weights_files : str
            Path to weights file
        """
        # TODO: these two methods are basically the same
        if not self.initialised:
            self.initialise()
        self.model.load_state_dict(torch.load(weights_file))
        self.model.eval()
        self.weights_file = weights_file

    def reload_weights(self, weights_file):
        """
        Trys to the load the weights file and if not, trys to load
        the weights file stored internally.

        Parameters
        ----------
        weights_files : str
            Path to weights file
        """
        if weights_file is None:
            weights_file = self.weights_file
        logger.debug(f'Reloading weights from {weights_file}')
        self.load_weights(weights_file)

    def reset_model(self, weights=True, permutations=False):
        """
        Reset the weights of the model and optimiser.

        Parameters
        ----------
        weights : bool, optional
            If true the model weights are reset.
        permutations : bool, optional
            If true any permutations (linear transforms) are reset.
        """
        if not any([weights, permutations]):
            logger.debug('Nothing to reset')
            return
        if weights:
            self.model.apply(reset_weights)
            logger.debug('Reset weights')
        elif permutations:
            self.model.apply(reset_permutations)
            logger.debug('Reset linear transforms')
        self.optimiser = self.get_optimiser()
        logger.debug('Reseting optimiser')

    def forward_and_log_prob(self, x):
        """
        Forward pass through the model and return the samples in the latent
        space with their log probabilties

        Parameters
        ----------
        x : ndarray
            Array of samples

        Returns
        -------
        z : ndarray
            Samples in the latent space
        log_prob : ndarray
            Log probabilties for each samples
        """
        x = torch.Tensor(x.astype(np.float32)).to(self.device)
        self.model.eval()
        with torch.no_grad():
            z, log_prob = self.model.forward_and_log_prob(x)

        z = z.detach().cpu().numpy()
        log_prob = log_prob.detach().cpu().numpy()
        return z, log_prob

    def sample_and_log_prob(self, N=1, z=None, alt_dist=None):
        """
        Generate samples from samples drawn from the base distribution or
        and alternative distribution from provided latent samples

        Parameters
        ----------
        N : int, optional
            Number of samples to draw if z is not specified
        z : ndarray, optional
            Array of latent samples to map the the data space, if ``alt_dist``
            is not specified they are assumed to be drawn from the base
            distribution of the flow.
        alt_dist : :obj:`nflows.distribution.Distribution`
            Distribution object from which the latent samples z were
            drawn from. Must have a ``log_prob`` method that accepts an
            instance of ``torch.Tensor``

        Returns
        -------
        samples : ndarray
            Array containing samples in the latent space.
        log_prob : ndarray
            Array containing the log probabaility that corresponds to each
            sample.
        """
        if self.model is None:
            raise RuntimeError('Model is not initialised yet!')
        if self.model.training:
            self.model.eval()
        if z is None:
            with torch.no_grad():
                x, log_prob = self.model.sample_and_log_prob(N)
        else:
            if alt_dist is not None:
                log_prob_fn = alt_dist.log_prob
            else:
                log_prob_fn = self.model.base_distribution_log_prob

            with torch.no_grad():
                if isinstance(z, np.ndarray):
                    z = torch.Tensor(z.astype(np.float32)).to(self.device)
                log_prob = log_prob_fn(z)
                x, log_J = self.model.inverse(z, context=None)
                log_prob -= log_J

        x = x.detach().cpu().numpy()
        log_prob = log_prob.detach().cpu().numpy()
        return x, log_prob

    def __getstate__(self):
        state = self.__dict__.copy()
        state['initialised'] = False
        del state['optimiser']
        del state['model']
        del state['model_config']
        return state



# -*- coding: utf-8 -*-
"""
Object for defining the use-defined model.
"""
import numpy as np

from .livepoint import (
        parameters_to_live_point,
        numpy_array_to_live_points,
        get_dtype,
        DEFAULT_FLOAT_DTYPE
        )


class Model:
    """Base class for the user-defined model being sampled.

    The user must define the attributes ``names`` ``bounds`` and the metods
    ``log_likelihood`` and ``log_prior``.

    The user can also define the reparemeterisations here instead of in
    the keyword arguments passed to the sampler.


    Attributes
    ----------
    names : list of str
        List of names of parameters, e.g. ['x', 'y']
    bounds : dict
        Dictionary of prior bounds, e.g. {'x': [-1, 1], 'y': [-1, 1]}
    reparameterisations : dict
        Dictionary of reparameterisations that overrides the values specified
        with keyword arguments.
    likelihood_evaluations : int
        Number of likelihood evaluations
    """

    names = []
    bounds = {}
    reparameterisations = None
    likelihood_evaluations = 0
    _lower = None
    _upper = None

    @property
    def dims(self):
        """Number of dimensions in the model"""
        if self.names:
            return len(self.names)
        else:
            return None

    @property
    def lower_bounds(self):
        """Lower bounds on the priors"""
        if self._lower is None:
            bounds_array = np.array(list(self.bounds.values()))
            self._lower = bounds_array[:, 0]
            self._upper = bounds_array[:, 1]
        return self._lower

    @property
    def upper_bounds(self):
        """Upper bounds on the priors"""
        if self._upper is None:
            bounds_array = np.array(list(self.bounds.values()))
            self._lower = bounds_array[:, 0]
            self._upper = bounds_array[:, 1]
        return self._upper

    def new_point(self, N=1):
        """
        Create a new LivePoint, drawn from within bounds.

        See `new_point_log_prob` if changing this method.

        Parameters
        ----------
        N : int, optional
            Number of points to draw. By default draws one point. If N > 1
            points are drawn using a faster method.

        Returns
        -------
        ndarray
            Numpy structured array with fields for each parameter
            and log-prior (logP) and log-likelihood (logL)
        """
        if N > 1:
            return self._multiple_new_points(N)
        else:
            return self._single_new_point()

    def new_point_log_prob(self, x):
        """
        Computes the proposal probabaility for a new point.

        This does not assume the that points will be drawn according to the
        prior. If `new_point` is redefined this method must be updated to
        match.

        Parameters
        ----------
        x : ndarray
            Points in a structured array

        Returns
        -------
        ndarray
            Log proposal probability for each point
        """
        return np.zeros(x.size)

    def _single_new_point(self):
        """
        Draw a single point within the prior

        Returns
        -------
        ndarray
            Numpy structured array with fields for each parameter
            and log-prior (logP) and log-likelihood (logL)
        """
        logP = -np.inf
        while (logP == -np.inf):
            p = parameters_to_live_point(
                    np.random.uniform(self.lower_bounds, self.upper_bounds,
                                      self.dims),
                    self.names)
            logP = self.log_prior(p)
        return p

    def _multiple_new_points(self, N):
        """
        Draw multiple points within the prior. Draws N points and accepts
        only those for which log-prior is finite.

        Parameters
        ----------
        N : int
            Number of points to draw

        Returns
        -------
        np.ndarray
            Numpy structured array with fields for each parameter
            and log-prior (logP) and log-likelihood (logL)
        """
        new_points = np.array([], dtype=get_dtype(self.names,
                                                  DEFAULT_FLOAT_DTYPE))
        while new_points.size < N:
            p = numpy_array_to_live_points(
                    np.random.uniform(self.lower_bounds, self.upper_bounds,
                                      [N, self.dims]),
                    self.names)
            logP = self.log_prior(p)
            new_points = np.concatenate([new_points, p[np.isfinite(logP)]])
        return new_points

    def log_prior(self, x):
        """
        Returns log-prior, must be defined by the user.
        """
        pass

    def log_likelihood(self, x):
        """
        Returns the log-likelihood, must be defined by the user.
        """
        pass

    def evaluate_log_likelihood(self, x):
        """
        Evaluate the log-likelihood and track the number of calls.

        Returns
        -------
        float
            Log-likelihood value

        """
        self.likelihood_evaluations += 1
        return self.log_likelihood(x)

    def verify_model(self):
        """
        Verify that the model is correctly setup. This includes checking
        the names, bounds and log-likelihood.
        """
        if not self.names:
            raise ValueError('Names for model parameters are not set')
        if not self.bounds:
            raise ValueError('Bounds are not set for model')

        for n in self.names:
            if n not in self.bounds.keys():
                raise RuntimeError(f'Missing bounds for {n}')

        logP = -np.inf
        counter = 0
        while (logP == -np.inf) or (logP == np.inf):
            x = numpy_array_to_live_points(
                    np.random.uniform(self.lower_bounds, self.upper_bounds,
                                      [1, self.dims]),
                    self.names)
            logP = self.log_prior(x)
            counter += 1
            if counter == 1000:
                raise RuntimeError('Could not draw valid point from within '
                                   'the prior after 10000 tries, check the '
                                   'log prior function.')

        if self.log_prior(x) is None:
            raise RuntimeError('Log-prior function did not return'
                               'a prior value')
        if self.log_likelihood(x) is None:
            raise RuntimeError('Log-likehood function did not return'
                               'a likelihood value')



# -*- coding: utf-8 -*-
"""
Specific priors for sampling gravitational-wave parameters with the x-prime
prior.
"""

import numpy as np


def log_uniform_prior(x, xmin=-1, xmax=1):
    """Unformalised log probability of uniform prior"""
    return np.log((x >= xmin) & (x <= xmax))


def log_2d_cartesian_prior(x, y, k=np.pi):
    """
    Log probability of Cartesian coordinates for a uniform distibution of
    angles on [0, k] and a radial component drawn from a chi distribution
    with two degrees of freedom.
    """
    return - np.log(k) - ((x ** 2 + y ** 2) / 2)


def log_2d_cartesian_prior_sine(x, y):
    """
    Log probability of Cartesian coordinates for a sine distibution of
    angles and a radial component drawn from a chi distribution
    with two degrees of freedom.
    """
    r = x ** 2 + y ** 2
    y[y < 0] = 0
    return np.log(y / 2) - 0.5 * np.log(r) - (r / 2)


def log_3d_cartesian_prior(x, y, z):
    """
    Log probability of 3d Cartesian coordinates for an isotropic distribution
    of angles and a radial component drawn from a chi distribution with three
    degrees of freedom.
    """
    return - 1.5 * np.log(2 * np.pi) - (x ** 2 + y ** 2 + z ** 2) / 2


def log_spin_prior(s1x, s1y, s1z, s2x, s2y, s2z, k1=0.99, k2=0.99):
    """
    Log probability of the prior on the components of spin vectors
    assume the distribution of a_i is uniform.
    """
    r1 = s1x ** 2 + s1y ** 2 + s1z ** 2
    r2 = s2x ** 2 + s2y ** 2 + s2z ** 2
    log_p = np.log(r1 <= (k1 ** 2)) + np.log(r2 <= (k2 ** 2))
    log_p += (-np.log(16 * np.pi ** 2 * k1 * k2) - np.log(r1) - np.log(r2))
    return log_p


def log_spin_prior_uniform(s1x, s1y, s1z, s2x, s2y, s2z, k1=0.99, k2=0.99):
    """
    Log probability of the prior on the components of spin vectors
    assume the distribution the components in uniform within the 2-ball.
    """
    r1 = s1x ** 2 + s1y ** 2 + s1z ** 2
    r2 = s2x ** 2 + s2y ** 2 + s2z ** 2
    log_p = np.log(r1 <= (k1 ** 2)) + np.log(r2 <= (k2 ** 2))
    log_p -= np.log(16 * np.pi ** 2 * k1 * k2 / 9)
    return log_p



# -*- coding: utf-8 -*-
"""
Implementation of Neural Spline Flows.
"""
import logging

import torch.nn.functional as F

from nflows.distributions import StandardNormal
from nflows import transforms
from nflows.nn.nets import ResidualNet
from nflows.utils import create_alternating_binary_mask

from .base import NFlow

logger = logging.getLogger(__name__)


class NeuralSplineFlow(NFlow):
    """
    Implementation of Neural Spline Flow

    See: https://arxiv.org/abs/1906.04032

    Parameters
    ----------
    features : int
        Number of features (dimensions) in the data space
    hidden_features : int
        Number of neurons per layer in each neural network
    num_layers : int
        Number of coupling tranformations
    num_blocks_per_layer : int
        Number of layers (or blocks for resnet) per nerual network for
        each coupling transform
    num_bins : int, optional (8)
        Number of bins to use for each spline
    activation : function
        Activation function implemented in torch
    dropout_probability : float, optional (0.0)
        Dropout probaiblity used in each layer of the neural network
    batch_norm_within_layers : bool, optional (False)
       Enable or disable batch norm within the neural network for each coupling
       transform
    batch_norm_between_layers : bool, optional (False)
       Enable or disable batch norm between coupling transforms
    linear_transform : {'permutaiton', 'lu', 'svd'}
        Linear transform to use between coupling layers. Not recommended when
        using a custom mask.
    kwargs : dict
        Additional kwargs parsed to the spline constructor, e.g. `tails` or
        `tail_bound`. See nflows for details
    """
    def __init__(
        self,
        features,
        hidden_features,
        num_layers,
        num_blocks_per_layer,
        num_bins=8,
        activation=F.relu,
        dropout_probability=0.0,
        batch_norm_within_layers=False,
        batch_norm_between_layers=False,
        apply_unconditional_transform=False,
        linear_transform='permutation',
        tails='linear',
        tail_bound=1.0,
        **kwargs
    ):
        if batch_norm_between_layers:
            logger.warning('BatchNorm cannot be used with splines')
            batch_norm_between_layers = False

        def create_linear_transform():
            if linear_transform == 'permutation':
                return transforms.RandomPermutation(features=features)
            elif linear_transform == 'lu':
                return transforms.CompositeTransform([
                    transforms.RandomPermutation(features=features),
                    transforms.LULinear(features, identity_init=True)
                ])
            elif linear_transform == 'svd':
                return transforms.CompositeTransform([
                    transforms.RandomPermutation(features=features),
                    transforms.SVDLinear(features, num_householder=10,
                                         identity_init=True)
                ])
            else:
                raise ValueError

        def create_resnet(in_features, out_features):
            return ResidualNet(
                in_features,
                out_features,
                hidden_features=hidden_features,
                num_blocks=num_blocks_per_layer,
                activation=activation,
                dropout_probability=dropout_probability,
                use_batch_norm=batch_norm_within_layers,
            )

        def spline_constructor(i):
            return transforms.PiecewiseRationalQuadraticCouplingTransform(
                mask=create_alternating_binary_mask(
                    features=features,
                    even=(i % 2 == 0)),
                transform_net_create_fn=create_resnet,
                num_bins=num_bins,
                apply_unconditional_transform=apply_unconditional_transform,
                tails=tails,
                tail_bound=tail_bound,
                **kwargs)

        transforms_list = []
        for i in range(num_layers):
            transforms_list.append(create_linear_transform())
            transforms_list.append(spline_constructor(i))

        distribution = StandardNormal([features])

        super().__init__(
            transform=transforms.CompositeTransform(transforms_list),
            distribution=distribution,
        )



# -*- coding: utf-8 -*-
"""
Test the RescaleToBound class.
"""
import numpy as np
import pytest

from nessai.reparameterisations import RescaleToBounds
from nessai.livepoint import get_dtype


@pytest.fixture()
def reparameterisation(model):
    def _get_reparameterisation(kwargs):
        return RescaleToBounds(parameters=model.names,
                               prior_bounds=model.bounds,
                               **kwargs)
    return _get_reparameterisation


@pytest.fixture(scope='function')
def assert_invertibility(model, n=100):
    def test_invertibility(reparam):
        x = model.new_point(N=n)
        x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
        log_j = 0

        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(
            x, x_prime, log_j)

        np.testing.assert_array_equal(x, x_re)

        x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))

        x_inv, x_prime_inv, log_j_inv = \
            reparam.inverse_reparameterise(x_in, x_prime_re, log_j)

        np.testing.assert_array_equal(x, x_inv)
        np.testing.assert_array_equal(x_prime_re, x_prime_inv)
        np.testing.assert_array_equal(log_j_re, -log_j_inv)

        return True

    return test_invertibility


@pytest.mark.parametrize('rescale_bounds', [None, [0, 1]])
def test_rescale_bounds(reparameterisation, assert_invertibility,
                        rescale_bounds):
    """Test the different options for rescale to bounds"""
    reparam = reparameterisation({'rescale_bounds': rescale_bounds})
    if rescale_bounds is None:
        rescale_bounds = {p: [-1, 1] for p in reparam.parameters}
    elif isinstance(rescale_bounds, list):
        rescale_bounds = {p: rescale_bounds for p in reparam.parameters}

    assert reparam.rescale_bounds == rescale_bounds
    assert assert_invertibility(reparam)


@pytest.mark.parametrize('boundary_inversion', [False,
                                                True,
                                                ['x']])
def test_boundary_inversion(reparameterisation, assert_invertibility,
                            boundary_inversion):
    """Test the different options for rescale to bounds"""
    reparam = reparameterisation({'boundary_inversion': boundary_inversion})

    assert assert_invertibility(reparam)



# -*- coding: utf-8 -*-
"""
Implementation of MaskedAutoregressiveFlow.
"""
from nflows.flows import MaskedAutoregressiveFlow as BaseMAF


class MaskedAutoregressiveFlow(BaseMAF):
    """
    Wrapper for MaskedAutoregressiveFlow included in nflowsi

    Adds the additional methods that are used in FlowModel.

    See: https://github.com/bayesiains/nflows/blob/master/nflows/flows/
    """

    def forward(self, x, context=None):
        """
        Apply the forward transformation and return samples in the latent
        space and log-Jacobian determinant.
        """
        return self._transform.forward(x)

    def inverse(self, z, context=None):
        """
        Apply the inverse transformation and return samples in the
        data space and log-Jacobian determinant.
        """
        return self._transform.inverse(z, context=context)

    def base_distribution_log_prob(self, z, context=None):
        """
        Computes the log probability of samples in the latent for
        the base distribution in the flow
        """
        return self._distribution.log_prob(z, context=context)

    def forward_and_log_prob(self, x, context=None):
        """
        Apply the forward transformation and compute the log probability
        of each sample

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the latent space
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        z, log_J = self.forward(x)
        log_prob = self.base_distribution_log_prob(z)
        return z, log_prob + log_J



# -*- coding: utf-8 -*-
"""
Specific proposal methods for sampling gravitational-wave models.
"""
import os
import logging

import numpy as np
from scipy.stats import chi, norm

from ..proposal import FlowProposal
from ..flowmodel import FlowModel
from ..utils import (
    replace_in_list,
    rescale_zero_to_one,
    inverse_rescale_zero_to_one,
    rescale_minus_one_to_one,
    inverse_rescale_minus_one_to_one,
    detect_edge)
from .utils import (
    determine_rescaled_bounds,
    angle_to_cartesian,
    cartesian_to_angle,
    zero_one_to_cartesian,
    cartesian_to_zero_one,
    transform_from_precessing_parameters,
    transform_to_precessing_parameters,
    rescale_and_logit,
    rescale_and_sigmoid,
    ComovingDistanceConverter)
from .priors import (
    log_uniform_prior,
    log_2d_cartesian_prior,
    log_2d_cartesian_prior_sine,
    log_3d_cartesian_prior,
    log_spin_prior,
    log_spin_prior_uniform)

from .reparameterisations import get_gw_reparameterisation

logger = logging.getLogger(__name__)


class LegacyGWFlowProposal(FlowProposal):
    """
    A proposal specific to gravitational wave CBC
    """
    def __init__(self, model, reparameterisations={}, **kwargs):
        super().__init__(model, **kwargs)

        self.set_reparameterisations(reparameterisations)
        # list to itnernally track reparemeterisations
        self._reparameterisations = []
        self._search_angles = {}
        self._inversion = {}
        self._log_inversion = {}
        self._log_radial = {}
        self._angle_conversion = {}
        self._rescaled_min = {}
        self._rescaled_max = {}

        self._x_prime_prior_parameters = \
            ['mass_ratio_inv', 'chirp_mass_prime', 'time', 'dc3',
             'sky_x', 'sky_y', 'sky_z', 'psi_x', 'psi_y', 'iota_x', 'iota_y',
             'tilt_1_x', 'tilt_1_y', 'tilt_2_x', 'tilt_2_y', 'a_1_x', 'a_1_y',
             'a_2_x', 'a_2_y', 'phi_12_x', 'phi_12_y', 'phi_jl_x', 'phi_jl_y',
             'theta_jn_x', 'theta_jn_y',
             's1x', 's1y', 's1z', 's2x', 's2y', 's2z']

        self._default_inversion_parameters = ['mass_ratio', 'a_1', 'a_2',
                                              'luminosity_distance']

        self._default_angles = ['psi', 'phase', 'iota', 'theta_jn', 'dec',
                                'ra', 'tilt_1', 'tilt_2', 'cos_theta_jn',
                                'cos_tilt_1', 'cos_tilt_2', 'phi_12', 'phi_jl']

    def set_reparameterisations(self, reparameterisations):
        """
        Set the relevant reparamterisation flags
        """
        defaults = dict(mass_inversion=False,
                        reduced_quaternions=True, distance_rescaling=False,
                        norm_quaternions=False, rescale_angles=True,
                        euler_convention='ZYZ', angular_decomposition=True,
                        minus_one_to_one=True, log_inversion=False,
                        log_radial=False, inversion=True, exclude=[],
                        convert_to_angle=False, default_rescaling=[],
                        spin_conversion=False, spin_conversion_config={},
                        uniform_distance_parameter=False,
                        uniform_distance_parameter_config={},
                        spin_logit=False,
                        spin_logit_config={},
                        use_x_prime_prior=False)
        defaults.update(reparameterisations)

        if defaults['mass_inversion'] and defaults['inversion']:
            raise RuntimeError(
                'Mass inversion and inversion are not compatible')

        if (defaults['inversion'] is True and
                defaults['log_inversion'] is True):
            raise RuntimeError('To use inversion and log-inversion, specify '
                               'the parameters for each as list')
        if (isinstance(defaults['inversion'], list) and
                defaults['log_inversion'] is True):
            raise RuntimeError('Cannot set inversion to a list and use '
                               'default parameters for log inversion.')
        if (isinstance(defaults['log_inversion'], list) and
                defaults['inversion'] is True):
            raise RuntimeError('Cannot set log inversion to a list and use '
                               'default parameters for inversion.')

        if all([isinstance(inv, list) for inv in [defaults['inversion'],
                                                  defaults['log_inversion']]]):
            s = set(defaults['inversion']) & set(defaults['log_inversion'])
            if s:
                raise RuntimeError('Inversion and log_inversion have common '
                                   f'parameters: {s}')

        logger.info('Reparameterisations:')
        for k, v in defaults.items():
            logger.info(f'{k}: {v}')
            setattr(self, k, v)

    def setup_angle(self, name, radial_name=False, scale=1.0, zero='bound'):
        """
        Add an angular parameter to the list of reparameterisations
        """
        if not radial_name:
            radial_name = name + '_radial'
            self.names.append(radial_name)
            self.rescaled_names.append(radial_name)

        x_name = name + '_x'
        y_name = name + '_y'
        replace_in_list(self.rescaled_names, [name, radial_name],
                        [x_name, y_name])

        self._search_angles[name] = {
            'angle': name, 'radial': radial_name,
            'x': x_name, 'y': y_name, 'scale': scale, 'zero': zero}

        logger.debug(f'Added {name} with config: {self._search_angles[name]}')

    @property
    def inversion_parameters(self):
        """
        Returns a list of parameters to which an inversion (normal or log)
        """
        parameters = []
        if isinstance(self.inversion, list):
            parameters += self.inversion
        if isinstance(self.log_inversion, list):
            parameters += self.log_inversion
        if not parameters and any([self.inversion, self.log_inversion]):
            parameters = self._default_inversion_parameters
        return parameters

    def add_inversion(self, name):
        """
        Setup inversion
        """
        rescaled_name = name + '_inv'
        replace_in_list(self.rescaled_names, [name],
                        [rescaled_name])

        # if name exists in the physical space, change it
        # else (e.g. radial parameters) leave as is
        self._inversion[name] = {
            'name': name, 'rescaled_name': rescaled_name,
            'rescale': True, 'invert': None,
            'min': self.model.bounds[name][0],
            'max': self.model.bounds[name][1]}

        logger.debug(f'Added {name} to parameters with inversion')

    def add_log_inversion(self, name):
        """
        Setup log inversion
        """
        # If upper bound is not 1, then rescale so require offset
        if not self.model.bounds[name][1] == 1 or \
                self.model.bounds[name][0] == 0:
            rescale = True
            offset = 0.0 * np.ptp(self.model.bounds[name])
        else:
            rescale = False
            offset = 0.0

        rescaled_name = name + '_inv'
        replace_in_list(self.rescaled_names, [name],
                        [rescaled_name])

        # if name exists in the physical space, change it
        # else (e.g. radial parameters) leave as is
        self._log_inversion[name] = {
            'name': name, 'rescaled_name': rescaled_name,
            'offset': offset, 'rescale': rescale, 'invert': None}

        logger.debug(f'Added {name} to parameters with log inversion')

    def add_angle_conversion(self, name, mode='split'):
        radial_name = name + '_radial'
        self.names.append(radial_name)
        self.rescaled_names.append(radial_name)

        x_name = name + '_x'
        y_name = name + '_y'
        replace_in_list(self.rescaled_names, [name, radial_name],
                        [x_name, y_name])

        self._angle_conversion[name] = {
            'name': name, 'radial': radial_name,
            'x': x_name, 'y': y_name, 'apply': True, 'mode': mode}

        logger.debug(f'{name} will be converted to an angle')

    def configure_time(self):
        """
        Configure the time parameter if present
        """
        time = [t for t in self.names if 'time' in t]
        if len(time) > 1:
            raise RuntimeError(f'Found more than one time: {time}')
        elif not len(time):
            self.time = False
        else:
            self.time = time[0]
            replace_in_list(self.rescaled_names, [self.time], ['time'])
            self._remaining.remove(self.time)
            # geocent time is handled different to other parameters,
            # we leave it in the defaults and change the prior bounds
            # we then only need to subtract the offset if it present
            # set offset as the midpoint of the prior
            # the bounds will then be +/- duration/2
            self.time_offset = self.model.bounds[self.time][0] \
                + np.ptp(self.model.bounds[self.time]) / 2
            logger.debug(f'Time offset: {self.time_offset}')
            # Save the bounds since we're using different bounds
            self.time_bounds = self.model.bounds[self.time] - self.time_offset
            self._rescaled_min['time'] = -1
            self._rescaled_max['time'] = 1

    def configure_sky(self):
        """
        Configure the sky parameters
        """
        if all(p in self.names for p in ['ra', 'dec']):
            from .utils import ra_dec_to_cartesian, cartesian_to_ra_dec
            self.sky_angles = ['ra', 'dec']
            self.sky_to_cartesian = ra_dec_to_cartesian
            self.cartesian_to_sky = cartesian_to_ra_dec
        elif all(p in self.names for p in ['azimuth', 'zenith']):
            from .utils import (azimuth_zenith_to_cartesian,
                                cartesian_to_azimuth_zenith)
            self.sky_angles = ['azimuth', 'zenith']
            self.sky_to_cartesian = azimuth_zenith_to_cartesian
            self.cartesian_to_sky = cartesian_to_azimuth_zenith
        elif any(p in self.names for p in
                 ['ra', 'dec', 'azimuth', 'zenith']):
            raise RuntimeError(
                'Cannot use angular decompoisiton with only'
                'one of the two sky angles')
        else:
            self.sky_angles = []

        if self.angular_decomposition and self.sky_angles:
            replace_in_list(self.rescaled_names,
                            self.sky_angles, ['sky_x', 'sky_y'])
            [self._remaining.remove(a) for a in self.sky_angles]
            if ('luminosity_distance' not in self.names or
                    'luminosity_distance' not in self._remaining or
                    'luminosity_distance' in self.default_rescaling):
                self.names.append('sky_radial')
                self.distance = 'sky_radial'
                self.rescaled_names.append('sky_z')
            else:
                self.distance = 'luminosity_distance'
                replace_in_list(self.rescaled_names, [self.distance],
                                ['sky_z'])
                self._remaining.remove('luminosty_disance')
            self._reparameterisations.append('sky')
            logger.info('Using angular decomposition of sky for: '
                        f'{self.sky_angles}')
        elif self.sky_angles:
            logger.warning(
                'Sampling sky but not using Cartesian reparmeterisation!')

    def configure_angles(self):
        """
        Configure angles
        """
        if self.angular_decomposition:
            if self.rescale_angles:
                if not isinstance(self.rescale_angles, list):
                    if isinstance(self.rescale_angles, bool):
                        self.rescale_angles = self._default_angles
                    elif isinstance(self.rescale_angles, str):
                        if self.rescale_angles == 'all':
                            self.rescale_angles = self._default_angles
                        else:
                            raise ValueError(
                                'Unknown value for rescale_angles: '
                                f'{self.rescale_angles}')
                logger.debug(f'Angles to rescale {self.rescale_angles}')
            else:
                self.rescale_angles = []

            logger.debug('Checking source angles')
            for a in ['psi', 'theta_jn', 'iota', 'phase', 'cos_theta_jn']:
                if a in self.names and a in self._remaining:
                    if a in self.rescale_angles:
                        scale = 2. * np.pi / np.ptp(self.model.bounds[a])
                    else:
                        scale = 1.0
                    if a in ['cos_theta_jn']:
                        zero = 'centre'
                    else:
                        zero = 'bound'
                    self.setup_angle(a, scale=scale, zero=zero)
                    self._remaining.remove(a)

            logger.debug('Checking spin angles')
            for i in [1, 2]:
                if f'tilt_{i}' in self.names:
                    a = f'tilt_{i}'
                elif f'cos_tilt_{i}' in self.names:
                    a = f'cos_tilt_{i}'
                else:
                    continue

                if a in self._remaining:
                    logger.debug(f'Adding reparameterisations for {a}')
                    radial = f'a_{i}'
                    if 'cos' in a:
                        zero = 'centre'
                    else:
                        zero = 'bound'
                    if a in self.rescale_angles:
                        scale = 2. * np.pi / np.ptp(self.model.bounds[a])
                    else:
                        scale = 1.0

                    if (radial in self.names and
                            radial in self._remaining and
                            not (radial in self._log_inversion or
                                 radial in self._inversion or
                                 radial in self._angle_conversion or
                                 radial in self.default_rescaling)):
                        self.setup_angle(a, radial, scale=scale, zero=zero)
                        self._remaining.remove(radial)
                    else:
                        self.setup_angle(a, scale=scale, zero=zero)
                    self._remaining.remove(a)

            for a in ['phi_jl', 'phi_12']:
                if a in self.names and a in self._remaining:
                    self.setup_angle(a, scale=1.0, zero='bound')
                    self._remaining.remove(a)
        else:
            logger.warning('Angles are not coverted to Cartesian!')

    def set_rescaling(self):
        """
        Set the rescaling functions
        """
        self.names = self.model.names.copy()
        self.rescaled_names = self.names.copy()

        self._remaining = self.names.copy()
        [self._remaining.remove(p) for p in self.default_rescaling]

        self._min = {n: self.model.bounds[n][0] for n in self.model.names}
        self._max = {n: self.model.bounds[n][1] for n in self.model.names}
        if isinstance(self.inversion_type, str):
            if self.inversion_type not in ('split', 'duplicate', 'reflexion'):
                raise RuntimeError(
                        f'Unknown inversion type: {self.inversion_type}')
            if self.inversion is True:
                self.inversion_type = \
                    {p: self.inversion_type for
                     p in self._default_inversion_parameters}
            elif self.log_inversion is True:
                self.inversion_type = \
                    {p: self.inversion_type for
                     p in self._default_inversion_parameters}
            else:
                self.inversion_type = \
                    {p: self.inversion_type for
                     p in self.inversion_parameters}

        elif isinstance(self.inversion_type, dict) and self.inversion_type:
            if self.inversion is True:
                self.inversion = list(self.inversion_type.keys())
            elif self.log_inversion is True:
                self.log_inversion = list(self.inversion_type.keys())
            elif (not self.inversion and not self.log_inversion):
                raise RuntimeError('Inversion type specified as dict but '
                                   'inverison and log inversion are both '
                                   'disabled')
            elif (set(self.inversion_parameters)
                  - set(self.inversion_type.keys())):
                raise RuntimeError('If inversion is a list and inversion '
                                   'type is a dict, the entries in the '
                                   'list must match the keys!')
        if self.log_inversion:
            if isinstance(self.log_inversion, list):
                for p in self.log_inversion:
                    self.add_log_inversion(p)
                    self._remaining.remove(p)
            else:
                for p in ['mass_ratio', 'luminosity_distance', 'a_1', 'a_2']:
                    if p in self.names:
                        self.add_log_inversion(p)
                        self._remaining.remove(p)

        if self.inversion:
            logger.info(f'Inversion types: {self.inversion_type}')
            if isinstance(self.inversion, list):
                for p in self.inversion:
                    if p in self.names:
                        self.add_inversion(p)
                        self._remaining.remove(p)
                    elif not p == 'dc3':
                        logger.debug(f'Cannot apply inversion to {p}, '
                                     'parameter not being sampled')
            else:
                for p in self._default_inversion_parameters:
                    if p in self.names:
                        self.add_inversion(p)
                        self._remaining.remove(p)
                    else:
                        logger.debug(f'Cannot apply inversion to {p}, '
                                     'parameter not being sampled')

        if self.log_radial:
            log_radial = ['luminosity_distance', 'a_1', 'a_2']
            if not isinstance(self.log_radial, list):
                self._log_radial = [p for p in log_radial
                                    if p not in self._log_inversion]
            else:
                self._log_radial = [p for p in self.log_radial
                                    if p not in self._log_inversion]
            [self._remaining.remove(p) for p in self._log_radial]
            logger.debug(f'Using log radial for {self._log_radial}')

        if self.uniform_distance_parameter:
            self.setup_uniform_distance_parameter(
                **self.uniform_distance_parameter_config)

        if self.spin_conversion:
            self.configure_spin_conversion(**self.spin_conversion_config)

        if self.convert_to_angle:
            if isinstance(self.convert_to_angle, list):
                for p in self.convert_to_angle:
                    if p in self.names:
                        self.add_angle_conversion(p, mode='split')
                        self._remaining.remove(p)
                    else:
                        logger.debug(f'Cannot convert {p} to angle, '
                                     'parameter not being sampled')
            if isinstance(self.convert_to_angle, dict):
                for k, v in self.convert_to_angle.items():
                    if k in self.names:
                        self.add_angle_conversion(k, mode=v)
                        self._remaining.remove(k)
                    else:
                        logger.debug(f'Cannot convert {k} to angle, '
                                     'parameter not being sampled')

        else:
            self.convert_to_angle = []

        if self.spin_logit:
            self.setup_spin_logit(**self.spin_logit_config)

        self.configure_sky()

        self.configure_time()

        self.configure_angles()

        if self.mass_inversion:
            raise NotImplementedError()

        # Default -1 to 1 rescaling
        if self.minus_one_to_one:
            self.default_rescaling += list(
                (set(self.names) & set(self.rescaled_names))
                - set(self.exclude) - set(self.default_rescaling))
            replace_in_list(self.rescaled_names, self.default_rescaling,
                            [d + '_prime' for d in self.default_rescaling])

        self._rescale_factor = np.ptp(self.rescale_bounds)
        self._rescale_shift = self.rescale_bounds[0]

        if (not all(p in self._x_prime_prior_parameters
                    for p in self.rescaled_names)
                and self.use_x_prime_prior):
            raise RuntimeError(
                'x prime space includes parameters that are not included in x '
                f'prime priors.\n x prime parameters: {self.rescaled_names}\n'
                'x prime parameters in prior: ',
                self._x_prime_prior_parameters)

        self.rescale_parameters = 'all'
        logger.info(f'x space parameters: {self.names}')
        logger.info(f'parameters to rescale {self.rescale_parameters}')
        logger.info(f'x prime space parameters: {self.rescaled_names}')

    def check_state(self, x):
        """
        Check the state of the rescaling before training
        """
        if self._log_inversion:
            for c in self._log_inversion.values():
                c['invert'] = None
        if self._inversion:
            for c in self._inversion.values():
                c['invert'] = None
                if self.update_bounds:
                    c['min'] = np.min(x[c['name']])
                    c['max'] = np.max(x[c['name']])

        if self.uniform_distance_parameter:
            self._dc3_invert = None
            if self.update_bounds:
                self._dc3_min = \
                    self.convert_to_dc3(np.min(x['luminosity_distance']))
                self._dc3_max = \
                    self.convert_to_dc3(np.max(x['luminosity_distance']))

        if self.update_bounds:
            self._min = {n: np.min(x[n]) for n in self.model.names}
            self._max = {n: np.max(x[n]) for n in self.model.names}

        if self.use_x_prime_prior:
            self.update_rescaled_bounds()

    def update_rescaled_bounds(self, rescaled_names=None,
                               xmin=None, xmax=None):
        if rescaled_names is not None:
            for rn, mn, mx in zip(rescaled_names, xmin, xmax):
                self._rescaled_min[rn] = xmin
                self._rescaled_max[rn] = xmax

        else:
            for n, rn in zip(['chirp_mass'],
                             ['chirp_mass_prime']):
                if n in self.model.names:
                    self._rescaled_min[rn], _ = rescale_minus_one_to_one(
                        self.model.bounds[n][0], self._min[n], self._max[n])
                    self._rescaled_max[rn], _ = rescale_minus_one_to_one(
                        self.model.bounds[n][1], self._min[n], self._max[n])

    def setup_uniform_distance_parameter(self, scale_factor=1000, **kwargs):
        """
        Set up the uniform distance parameter dc3

        Parameters
        ----------
        scale_factor : float, (optional)
            Factor used to rescale comoving distance
        kwargs :
            Keyword arguments parsed to `ComovingDistanceConverter`
        """
        if 'luminosity_distance' not in self.names:
            raise RuntimeError('Uniform distance parameter is only compatible '
                               'with luminosity distance')
        if not self.use_x_prime_prior:
            raise RuntimeError('Cannot use dc3 without x prime prior')

        self.distance_converter = ComovingDistanceConverter(
            d_min=self.model.bounds['luminosity_distance'][0],
            d_max=self.model.bounds['luminosity_distance'][1],
            **kwargs)

        self._dc3_prior_min = \
            self.convert_to_dc3(self.model.bounds['luminosity_distance'][0])
        self._dc3_prior_max = \
            self.convert_to_dc3(self.model.bounds['luminosity_distance'][1])

        self._dc3_min = self._dc3_prior_min.copy()
        self._dc3_max = self._dc3_prior_max.copy()

        replace_in_list(self.rescaled_names, ['luminosity_distance'], ['dc3'])
        self._remaining.remove('luminosity_distance')

    def convert_to_dl(self, dc3):
        """
        Convert from uniform distance parameter dc3 to luminosity distance
        """
        return self.distance_converter.from_uniform_parameter(dc3)[0]

    def convert_to_dc3(self, dl):
        """
        Convert to uniform distance parameter dc3
        """
        return self.distance_converter.to_uniform_parameter(dl)[0]

    def setup_spin_logit(self, fuzz_factor=0.01):
        if 'a_1' in self.names and 'a_2' in self.names:
            self._spin_fuzz_factor = fuzz_factor
            self._remaining.remove('a_1')
            self._remaining.remove('a_2')
            replace_in_list(self.rescaled_names, ['a_1', 'a_2'],
                            ['a_1_logit', 'a_2_logit'])
        else:
            logger.debug('Missing spin magnitudes')

    def configure_spin_conversion(self, m1=20, m2=20, phase=0, f_ref=20,
                                  scale_factor=1, use_cbrt=True):
        self._m1 = float(m1)
        self._m2 = float(m2)
        self._phase = float(phase)
        self._f_ref = float(f_ref)
        self.scale_factor = scale_factor
        self._spin_use_cbrt = use_cbrt

        if self._spin_use_cbrt:
            self._spin_k1 = np.cbrt(self.model.bounds['a_1'][1])
            self._spin_k2 = np.cbrt(self.model.bounds['a_2'][1])
        else:
            self._spin_k1 = self.model.bounds['a_1'][1]
            self._spin_k2 = self.model.bounds['a_2'][1]

        self.precessing_params = \
            ['theta_jn', 'phi_jl', 'tilt_1', 'tilt_2', 'phi_12', 'a_1', 'a_2']
        self.cartesian_spin_params = \
            ['iota', 's1x', 's1y', 's1z', 's2x', 's2y', 's2z']
        replace_in_list(self.rescaled_names,
                        self.precessing_params,
                        self.cartesian_spin_params)

        [self._remaining.remove(p) for p in self.precessing_params]

        self.names.append('iota_radial')
        self.rescaled_names.append('iota_radial')

        replace_in_list(self.rescaled_names, ['iota', 'iota_radial'],
                        ['iota_x', 'iota_y'])

        if 'iota' in self.rescale_angles:
            self._iota_scale = 2.0
        else:
            self._iota_scale = 1.0

    def _convert_spins_from_precessing(self, theta_jn, phi_jl, tilt_1,
                                       tilt_2, phi_12, a_1, a_2, m1=None,
                                       m2=None, f_ref=None, phase=None):
        if m1 is None:
            m1 = self._m1
        if m2 is None:
            m2 = self._m2
        if f_ref is None:
            f_ref = self._f_ref
        if phase is None:
            phase = self._phase

        log_J = 0

        if self._spin_use_cbrt:
            a_1 = np.cbrt(a_1)
            a_2 = np.cbrt(a_2)
            log_J += (-np.log(3) - 2 * np.log(a_1))
            log_J += (-np.log(3) - 2 * np.log(a_2))

        iota, s1x, s1y, s1z, s2x, s2y, s2z, lj = \
            transform_from_precessing_parameters(
                theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2,
                m1, m2, f_ref, phase)

        log_J += lj

        s1x *= self.scale_factor
        s1y *= self.scale_factor
        s1z *= self.scale_factor
        s2x *= self.scale_factor
        s2y *= self.scale_factor
        s2z *= self.scale_factor

        return iota, s1x, s1y, s1z, s2x, s2y, s2z, log_J

    def _convert_spins_to_precessing(self, iota, s1x, s1y, s1z, s2x, s2y,
                                     s2z, m1=None, m2=None, f_ref=None,
                                     phase=None):
        if m1 is None:
            m1 = self._m1
        if m2 is None:
            m2 = self._m2
        if f_ref is None:
            f_ref = self._f_ref
        if phase is None:
            phase = self._phase

        s1x /= self.scale_factor
        s1y /= self.scale_factor
        s1z /= self.scale_factor
        s2x /= self.scale_factor
        s2y /= self.scale_factor
        s2z /= self.scale_factor
        log_J = 0

        theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2, lj = \
            transform_to_precessing_parameters(
                iota, s1x, s1y, s1z, s2x, s2y, s2z, m1, m2, f_ref, phase)
        log_J += lj

        if self._spin_use_cbrt:
            log_J += (np.log(3) + 2 * np.log(a_1))
            log_J += (np.log(3) + 2 * np.log(a_2))
            a_1 = a_1 ** 3
            a_2 = a_2 ** 3

        return theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2, log_J

    def _apply_inversion(self, x, x_prime, log_J, name, rescaled_name,
                         invert, rescale, compute_radius, inversion_type,
                         xmin=None, xmax=None, x_array=None):

        # Allow specifying an array
        # This allows for an inverison to applied after other rescaling
        if x_array is None:
            x_array = x[name]
        if rescale:
            x_prime[rescaled_name], lj = rescale_zero_to_one(
                x_array, xmin=xmin, xmax=xmax)
            log_J += lj
        else:
            x_prime[rescaled_name] = x_array

        if invert == 'upper':
            x_prime[rescaled_name] = \
                    1 - x_prime[rescaled_name]

        if invert == 'both':
            if compute_radius:
                if x_prime[rescaled_name][0] < 0.5:
                    lower = np.arange(x.size, 2 * x.size)
                    upper = np.array([], dtype=int)
                else:
                    lower = np.array([], dtype=int)
                    upper = np.arange(x.size, 2 * x.size)
                x_prime = np.tile(x_prime, 2)
                x = np.tile(x, 2)
                log_J = np.tile(log_J, 2)
            else:
                lower_samples = \
                    np.where(x_prime[rescaled_name] <= 0.5)[0]
                upper_samples = \
                    np.where(x_prime[rescaled_name] > 0.5)[0]
                lower = np.random.choice(lower_samples,
                                         lower_samples.size // 2,
                                         replace=False)
                upper = np.random.choice(upper_samples,
                                         upper_samples.size // 2,
                                         replace=False)
            x_prime[rescaled_name][lower] *= -1
            x_prime[rescaled_name][upper] = \
                2.0 - x_prime[rescaled_name][upper]

        else:
            if inversion_type == 'duplicate' or compute_radius:
                x_inv = x_prime.copy()
                x_inv[rescaled_name] *= -1
                x_prime = np.concatenate([x_prime, x_inv])
                x = np.concatenate([x,  x])
                log_J = np.concatenate([log_J, log_J])
            else:
                inv = np.random.choice(x_prime.size,
                                       x_prime.size // 2,
                                       replace=False)
                x_prime[rescaled_name][inv] *= -1

        return x, x_prime, log_J

    def _reverse_inversion(self, x, x_prime, log_J, name, rescaled_name,
                           invert, rescale, xmin=None, xmax=None):

        if invert == 'both':
            lower = x_prime[rescaled_name] < 0.
            upper = x_prime[rescaled_name] > 1.
            x[name] = x_prime[rescaled_name]
            x[name][lower] *= -1
            x[name][upper] = 2 - x[name][upper]

        else:
            inv = x_prime[rescaled_name] < 0.
            x[name][~inv] = x_prime[rescaled_name][~inv]
            x[name][inv] = -x_prime[rescaled_name][inv]

            if invert == 'upper':
                x[name] = 1 - x[name]

        if rescale:
            x[name], lj = \
                inverse_rescale_zero_to_one(x[name],
                                            xmin=xmin,
                                            xmax=xmax)
            log_J += lj

        return x, x_prime, log_J

    def rescale(self, x, compute_radius=False, test=None):
        """
        Rescale from the x space to the x prime space
        """
        x_prime = np.zeros([x.size], dtype=self.x_prime_dtype)
        log_J = np.zeros(x_prime.size)

        x_prime['logP'] = x['logP']
        x_prime['logL'] = x['logL']

        if x.size == 1:
            x = np.array([x], dtype=x.dtype)

        if self.default_rescaling:
            for n in self.default_rescaling:
                x_prime[n + '_prime'] = self._rescale_factor \
                             * ((x[n] - self._min[n])
                                / (self._max[n] - self._min[n])) \
                             + self._rescale_shift

                log_J += (-np.log(self._max[n] - self._min[n])
                          + np.log(self._rescale_factor))

        if self.exclude:
            for n in self.exclude:
                if n in x.dtype.names:
                    x_prime[n] = x[n]

        if self.spin_conversion:
            precessing_params = [x[p] for p in self.precessing_params]
            cart_params = \
                self._convert_spins_from_precessing(*precessing_params)
            # Skip iota (first)
            log_J += cart_params[-1]
            for n, p in zip(self.cartesian_spin_params[1:], cart_params[1:-1]):
                x_prime[n] = p
            x_prime['iota_x'], x_prime['iota_y'], lj = angle_to_cartesian(
                cart_params[0], scale=self._iota_scale)
            log_J += lj

        if self.spin_logit:
            for a in ['a_1', 'a_2']:
                x_prime[a + '_logit'], lj = rescale_and_logit(
                    x[a],
                    self.model.bounds[a][0] - self._spin_fuzz_factor,
                    self.model.bounds[a][1] + self._spin_fuzz_factor)
                log_J += lj

        if self._log_inversion:
            for c in self._log_inversion.values():
                if c['rescale']:
                    x_prime[c['rescaled_name']], lj = rescale_zero_to_one(
                        x[c['name']],
                        xmin=self.model.bounds[c['name']][0] - c['offset'],
                        xmax=self.model.bounds[c['name']][1])
                    log_J += lj

                if c['invert'] is None:
                    c['invert'] = detect_edge(
                        x_prime[c['rescaled_name']],
                        **self.detect_edges_kwargs)

                if c['invert'] == 'lower':
                    x_prime[c['rescaled_name']] = \
                            1 - x_prime[c['rescaled_name']]

                x_prime[c['rescaled_name']] = \
                    np.log(x_prime[c['rescaled_name']])

                log_J -= x_prime[c['rescaled_name']]

                if c['invert']:

                    if self.inversion_type[c['name']] == 'duplicate':
                        x_inv = x_prime.copy()
                        x_inv[c['rescaled_name']] *= -1
                        x_prime = np.concatenate([x_prime, x_inv])
                        x = np.concatenate([x,  x])
                        log_J = np.concatenate([log_J, log_J])
                    else:
                        inv = np.random.choice(x_prime.size, x_prime.size // 2,
                                               replace=False)
                        x_prime[c['rescaled_name']][inv] *= -1

        if self.uniform_distance_parameter:
            dc3 = self.convert_to_dc3(x['luminosity_distance'])
            # Edge detection
            if self._dc3_invert is None:
                if 'dc3' in self.inversion:
                    self._dc3_invert = detect_edge(
                            dc3,
                            x_range=[self._dc3_prior_min, self._dc3_prior_max],
                            allow_both=False,
                            allowed_bounds=['upper'],
                            test=test,
                            **self.detect_edges_kwargs)
                else:
                    self._dc3_invert = False

                xmin, xmax = determine_rescaled_bounds(
                        self._dc3_prior_min,
                        self._dc3_prior_max,
                        self._dc3_min,
                        self._dc3_max,
                        self._dc3_invert)
                self.update_rescaled_bounds(['dc3'], [xmin], [xmax])

            if self._dc3_invert:
                x, x_prime, log_J = self._apply_inversion(
                    x, x_prime, log_J, 'luminosity_distance', 'dc3',
                    self._dc3_invert, True, compute_radius,
                    self.inversion_type['dc3'], xmin=self._dc3_min,
                    xmax=self._dc3_max, x_array=dc3)

            else:
                x_prime['dc3'], lj = rescale_minus_one_to_one(
                    dc3, self._dc3_min, self._dc3_max)
                log_J += lj

        if self._inversion:
            for c in self._inversion.values():
                if c['invert'] is None:
                    if self.inversion_type[c['name']] == 'reflexion':
                        both = True
                    else:
                        both = False
                    c['invert'] = detect_edge(
                        x[c['name']],
                        allow_both=both,
                        test=test,
                        **self.detect_edges_kwargs)
                    logger.debug(f"Inversion for {c['name']}: {c['invert']}")
                    if self.use_x_prime_prior:
                        # Set the prior bounds in the x_prime space
                        xmin, xmax = determine_rescaled_bounds(
                            self.model.bounds[c['name']][0],
                            self.model.bounds[c['name']][1],
                            c['min'],
                            c['max'],
                            invert=c['invert'])
                        self.update_rescaled_bounds([c['rescaled_name']],
                                                    [xmin], [xmax])
                if c['invert']:
                    x, x_prime, log_J = self._apply_inversion(
                        x, x_prime, log_J, c['name'], c['rescaled_name'],
                        c['invert'], c['rescale'], compute_radius,
                        self.inversion_type[c['name']], xmin=c['min'],
                        xmax=c['max'])
                else:
                    if c['rescale']:
                        x_prime[c['rescaled_name']], lj = \
                            rescale_minus_one_to_one(x[c['name']],
                                                     xmin=c['min'],
                                                     xmax=c['max'])
                        log_J += lj
                    else:
                        x_prime[c['rescaled_name']] = x[c['name']].copy()

        if self._angle_conversion:
            for c in self._angle_conversion.values():
                p, lj = rescale_zero_to_one(
                    x[c['name']],
                    xmin=self.model.bounds[c['name']][0],
                    xmax=self.model.bounds[c['name']][1])
                log_J += lj
                # if computing the radius, set duplicate=True
                if ((c['mode'] == 'split' and compute_radius)
                        or c['mode'] == 'duplicate'):
                    x_prime = np.concatenate([x_prime, x_prime])
                    x = np.concatenate([x,  x])
                    log_J = np.concatenate([log_J, log_J])
                    x_prime[c['x']], x_prime[c['y']], lj = \
                        zero_one_to_cartesian(p, mode='duplicate')
                    log_J += lj

                else:
                    x_prime[c['x']], x_prime[c['y']], lj = \
                        zero_one_to_cartesian(p, mode=c['mode'])

                    log_J += lj

        if 'sky' in self._reparameterisations:
            if self.distance == 'luminosity_distance':
                if 'luminosity_distance' in self._log_radial:
                    r, lj = rescale_zero_to_one(
                        x[self.distance],
                        xmin=self.model.bounds[self.distance][0] - 0.1,
                        xmax=self.model.bounds[self.distance][1])
                    log_J += lj
                    r = -np.log(r)
                    # logJ = log(1/r) where r is the value before applying log
                    # have log(1/r) so use this
                    log_J += r
                else:
                    r, lj = rescale_zero_to_one(
                        x[self.distance],
                        xmin=self.model.bounds[self.distance][0],
                        xmax=self.model.bounds[self.distance][1])
                    log_J += lj
            else:
                r = None

            x_prime['sky_x'], x_prime['sky_y'], x_prime['sky_z'], lj = \
                self.sky_to_cartesian(x[self.sky_angles[0]],
                                      x[self.sky_angles[1]], r)
            log_J += lj

        if self.time:
            t = x[self.time] - self.time_offset
            x_prime['time'], lj = rescale_minus_one_to_one(
                t, self.time_bounds[0], self.time_bounds[1])
            log_J += lj

        if self._search_angles:
            for a in self._search_angles.values():
                # if the radial parameter is present in x
                # use it, else samples with be drawn from a chi with
                # 2 d.o.f
                if a['radial'] in self.model.names:
                    r, lj = rescale_zero_to_one(
                        x[n],
                        xmin=self._min[a['radial']],
                        xmax=self._max[a['radial']])
                    log_J += lj
                    # r = np.log(r)
                    # log_J -= r
                    # r = 1 - r
                    # r = - np.log(1 - x[n])
                    # log_J += np.positive(r)   # log|J| = np.log(1-r)
                else:
                    r = None
                x_prime[a['x']], x_prime[a['y']], lj = angle_to_cartesian(
                    x[a['angle']], r=r, scale=a['scale'])

                log_J += lj

        if self.mass_inversion:
            raise NotImplementedError

        return x_prime, log_J

    def inverse_rescale(self, x_prime):
        """
        Rescale from the x prime  space to the x space
        """
        x = np.zeros([x_prime.size], dtype=self.x_dtype)
        log_J = np.zeros(x_prime.size)

        x['logP'] = x_prime['logP']
        x['logL'] = x_prime['logL']

        # Sort mass ratio first so that phase, tilt angles and magntiude
        # are correct before applying other rescaling
        if self.mass_inversion:
            raise NotImplementedError

        if 'sky' in self._reparameterisations:
            x[self.sky_angles[0]], x[self.sky_angles[1]], r, lj = \
                self.cartesian_to_sky(x_prime['sky_x'], x_prime['sky_y'],
                                      x_prime['sky_z'])
            log_J += lj

            if self.distance == 'luminosity_distance':
                if 'luminosity_distance' in self._log_radial:
                    log_J -= r.copy()
                    r = np.exp(-r)
                    r, lj = inverse_rescale_zero_to_one(
                        r, xmin=self.model.bounds[self.distance][0] - 0.1,
                        xmax=self.model.bounds[self.distance][1])
                    log_J += lj
                else:
                    r, lj = inverse_rescale_zero_to_one(
                        r, xmin=self.model.bounds[self.distance][0],
                        xmax=self.model.bounds[self.distance][1])
                    log_J += lj
            x[self.distance] = r

        if self.time:
            t, lj = inverse_rescale_minus_one_to_one(x_prime['time'],
                                                     self.time_bounds[0],
                                                     self.time_bounds[1])
            # This will break with casting rules
            x[self.time] = np.float64(t) + np.float64(self.time_offset)
            log_J += lj

        if self._search_angles:
            for a in self._search_angles.values():
                x[a['angle']], r, lj = cartesian_to_angle(
                    x_prime[a['x']], x_prime[a['y']], scale=a['scale'],
                    zero=a['zero'])
                log_J += lj
                # if the radial parameter is defined in the model
                # rescale it using the bounds
                if a['radial'] in self.model.names:
                    r, lj = inverse_rescale_zero_to_one(
                        r,
                        xmin=self._min[a['radial']],
                        xmax=self._max[a['radial']])
                    log_J += lj
                x[a['radial']] = r

        if self._angle_conversion:
            for c in self._angle_conversion.values():
                p, r, lj = cartesian_to_zero_one(
                    x_prime[c['x']], x_prime[c['y']])
                log_J += lj

                x[c['name']], lj = inverse_rescale_zero_to_one(
                    p, xmin=self.model.bounds[c['name']][0],
                    xmax=self.model.bounds[c['name']][1])
                log_J += lj
                x[c['radial']] = r

        if self._log_inversion:
            for c in self._log_inversion.values():
                inv = x_prime[c['rescaled_name']] > 0.
                x[c['name']][~inv] = np.exp(x_prime[c['rescaled_name']][~inv])
                x[c['name']][inv] = np.exp(-x_prime[c['rescaled_name']][inv])

                if c['invert'] == 'lower':
                    x[c['name']] = 1 - x[c['name']]

                # for q_inv < 0 conversion is
                # exp(q_inv) for q_inv > 0 exp(-q_inv)
                # so Jacobian is log(exp(+/-q_inv))
                # i.e. q_inv and - q_inv respectively
                log_J[~inv] += x_prime[c['rescaled_name']][~inv]
                log_J[inv] -= x_prime[c['rescaled_name']][inv]

                if c['rescale']:
                    x[c['name']], lj = inverse_rescale_zero_to_one(
                            x[c['name']],
                            xmin=self.model.bounds[c['name']][0] - c['offset'],
                            xmax=self.model.bounds[c['name']][1])
                    log_J += lj

        if self._inversion:
            for c in self._inversion.values():
                if c['invert']:
                    x, x_prime, log_J = self._reverse_inversion(
                            x, x_prime, log_J, c['name'], c['rescaled_name'],
                            c['invert'], c['rescale'],
                            xmin=c['min'], xmax=c['max'])
                else:
                    if c['rescale']:
                        x[c['name']], lj = inverse_rescale_minus_one_to_one(
                            x_prime[c['rescaled_name']],
                            xmin=c['min'], xmax=c['max'])
                        log_J += lj
                    else:
                        x[c['name']] = x_prime[c['rescaled_name']].copy()

        if self.spin_conversion:
            iota, x['iota_radial'], lj = cartesian_to_angle(
                x_prime['iota_x'], x_prime['iota_y'], scale=self._iota_scale)
            log_J += lj
            # iota is not present in x_prime
            cart_params = [x_prime[p] for p in self.cartesian_spin_params[1:]]
            precessing_params = \
                self._convert_spins_to_precessing(iota, *cart_params)
            for n, p in zip(self.precessing_params, precessing_params[:-1]):
                x[n] = p
            log_J += precessing_params[-1]

        if self.spin_logit:
            for a in ['a_1', 'a_2']:
                x[a], lj = rescale_and_sigmoid(
                    x_prime[a + '_logit'],
                    self.model.bounds[a][0] - self._spin_fuzz_factor,
                    self.model.bounds[a][1] + self._spin_fuzz_factor)

                log_J += lj

        if self.uniform_distance_parameter:
            if self._dc3_invert:
                x, x_prime, log_J = self._reverse_inversion(
                        x, x_prime, log_J, 'luminosity_distance', 'dc3',
                        self._dc3_invert, True,
                        xmin=self._dc3_min, xmax=self._dc3_max)
            else:
                x['luminosity_distance'], lj = \
                    inverse_rescale_minus_one_to_one(x_prime['dc3'],
                                                     xmin=self._dc3_min,
                                                     xmax=self._dc3_max)
                log_J += lj
            # Another step is requires to covert back to dL
            x['luminosity_distance'] = \
                self.convert_to_dl(x['luminosity_distance'])

        if self.default_rescaling:
            for n in self.default_rescaling:
                x[n] = (self._max[n] - self._min[n]) \
                    * (x_prime[n + '_prime'] - self._rescale_shift) \
                    / self._rescale_factor + self._min[n]

                log_J += (np.log(self._max[n] - self._min[n])
                          - np.log(self._rescale_factor))

        if self.exclude:
            for n in self.exclude:
                if n in x_prime.dtype.names:
                    x[n] = x_prime[n]

        return x, log_J

    def log_prior(self, x):
        """
        Modified log prior that handles radial parameters
        """
        log_p = self.model.log_prior(x[self.model.names])

        if 'sky' in self._reparameterisations:
            if self.distance == 'sky_radial':
                log_p += chi.logpdf(x[self.distance], 3)
        if self._search_angles:
            for a in self._search_angles.values():
                if not a['radial'] in self.model.names:
                    log_p += chi.logpdf(x[a['radial']], 2)

        if self._angle_conversion:
            for a in self._angle_conversion.values():
                log_p += chi.logpdf(x[a['radial']], 2)

        if self.spin_conversion and 'iota_radial' in self.names:
            log_p += chi.logpdf(x['iota_radial'], 2)

        return log_p

    def compute_rescaled_bounds(self, name):
        xmin = ((self.model.bounds[name][0] - self._min[name])
                / (self._max[name] - self._min[name]))

        xmin = min(xmin, 0)

        xmax = ((self.model.bounds[name][1] - self._min[name])
                / (self._max[name] - self._min[name]))

        xmax = max(xmax, 1)
        if name not in self.inversion or not self._inversion[name]['invert']:
            return 2 * xmin - 1, 2 * xmax - 1
        elif self._inversion[name]['invert'] == 'upper':
            return xmin - 1, 1 - xmin
        elif self._inversion[name]['invert'] == 'lower':
            return -xmax, xmax
        else:
            raise NotImplementedError

    def x_prime_log_prior(self, x_prime):
        """
        Priors redefined in the x_prime space

        Priors
        ------
        * Chirp mass and mass ratio: uniform (i.e. constant)
        * Time: uniform
        * polarsiation: uniform on pi (includes radial)
        """
        log_p = 0
        log_J = 0
        for n in ['chirp_mass_prime', 'mass_ratio_inv', 'time', 'dc3']:
            if n in self.rescaled_names:
                log_p += log_uniform_prior(x_prime[n],
                                           xmin=self._rescaled_min[n],
                                           xmax=self._rescaled_max[n])
        # Sky
        if 'sky_x' in self.rescaled_names:
            log_p += log_3d_cartesian_prior(x_prime['sky_x'], x_prime['sky_y'],
                                            x_prime['sky_z'])

        if 'psi_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior(
                x_prime['psi_x'], x_prime['psi_y'],
                k=self._search_angles['psi']['scale'] * np.pi)

        if 'phi_12_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior(
                x_prime['phi_12_x'], x_prime['phi_12_y'],
                k=self._search_angles['phi_12']['scale'] * np.pi)

        if 'phi_jl_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior(
                x_prime['phi_jl_x'], x_prime['phi_jl_y'],
                k=self._search_angles['phi_jl']['scale'] * np.pi)

        if 'a_1_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior(
                x_prime['a_1_x'], x_prime['a_1_y'],
                k=0.99)

        if 'a_2_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior(
                x_prime['a_2_x'], x_prime['a_2_y'],
                k=0.99)

        if 'iota_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior_sine(
                x_prime['iota_x'], x_prime['iota_y'])

        if 'theta_jn_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior_sine(
                x_prime['theta_jn_x'], x_prime['theta_jn_y'])

        if 'tilt_1_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior_sine(
                x_prime['tilt_1_x'], x_prime['tilt_1_y'])

        if 'tilt_2_x' in self.rescaled_names:
            log_p += log_2d_cartesian_prior_sine(
                x_prime['tilt_2_x'], x_prime['tilt_2_y'])

        if 's1x' in self.rescaled_names:
            if self._spin_use_cbrt:
                log_p += log_spin_prior_uniform(
                    x_prime['s1x'] / self.scale_factor,
                    x_prime['s1y'] / self.scale_factor,
                    x_prime['s1z'] / self.scale_factor,
                    x_prime['s2x'] / self.scale_factor,
                    x_prime['s2y'] / self.scale_factor,
                    x_prime['s2z'] / self.scale_factor,
                    k1=self._spin_k1,
                    k2=self._spin_k2)

            else:
                log_p += log_spin_prior(
                    x_prime['s1x'] / self.scale_factor,
                    x_prime['s1y'] / self.scale_factor,
                    x_prime['s1z'] / self.scale_factor,
                    x_prime['s2x'] / self.scale_factor,
                    x_prime['s2y'] / self.scale_factor,
                    x_prime['s2z'] / self.scale_factor,
                    k1=self._spin_k1,
                    k2=self._spin_k2)
        return log_p - log_J


class AugmentedGWFlowProposal(LegacyGWFlowProposal):

    def __init__(self, model, augment_features=1, **kwargs):
        super().__init__(model, **kwargs)
        self.augment_features = augment_features

    def set_rescaling(self):
        super().set_rescaling()
        self.augment_names = [f'e_{i}' for i in range(self.augment_features)]
        self.names += self.augment_names
        self.rescaled_names += self.augment_names
        self.exclude += self.augment_names
        logger.info(f'augmented x space parameters: {self.names}')
        logger.info(f'parameters to rescale {self.rescale_parameters}')
        logger.info(
            f'augmented x prime space parameters: {self.rescaled_names}'
        )

    def initialise(self):
        """
        Initialise the proposal class
        """
        if not os.path.exists(self.output):
            os.makedirs(self.output, exist_ok=True)

        self.set_rescaling()

        m = np.ones(self.rescaled_dims)
        m[-self.augment_features:] = -1
        self.flow_config['model_config']['kwargs']['mask'] = m

        self.flow_config['model_config']['n_inputs'] = self.rescaled_dims

        self.flow = FlowModel(config=self.flow_config, output=self.output)
        self.flow.initialise()
        self.populated = False
        self.initialised = True

    def rescale(self, x):
        x_prime, log_J = super().rescale(x)
        if x_prime.size == 1:
            x_prime[self.augment_names] = self.augment_features * (0.,)
        else:
            for an in self.augment_names:
                x_prime[an] = np.random.randn(x_prime.size)
        return x_prime, log_J

    def augmented_prior(self, x):
        """
        Log guassian for augmented variables
        """
        logP = 0.
        for n in self.augment_names:
            logP += norm.logpdf(x[n])
        return logP

    def log_prior(self, x):
        """
        Compute the prior probability
        """
        logP = super().log_prior(x)
        logP += self.augmented_prior(x)
        return logP


class GWFlowProposal(FlowProposal):
    """Wrapper for FlowProposal that has defaults for CBC-PE"""
    aliases = {
        'chirp_mass': ('mass', None),
        'mass_ratio': ('mass_ratio', None),
        'ra': ('sky-ra-dec', ['DEC', 'dec', 'Dec']),
        'dec': ('sky-ra-dec', ['RA', 'ra']),
        'azimuth': ('sky-az-zen', ['zenith', 'zen', 'Zen', 'Zenith']),
        'zenith': ('sky-az-zen', ['azimuth', 'az', 'Az', 'Azimuth']),
        'theta_1': ('angle-sine', None),
        'theta_2': ('angle-sine', None),
        'tilt_1': ('angle-sine', None),
        'tilt_2': ('angle-sine', None),
        'theta_jn': ('angle-sine', None),
        'iota': ('angle-sine', None),
        'phi_jl': ('angle-2pi', None),
        'phi_12': ('angle-2pi', None),
        'phase': ('angle-2pi', None),
        'psi': ('angle-pi', None),
        'geocent_time': ('time', None),
        'a_1': ('to-cartesian', None),
        'a_2': ('to-cartesian', None),
        'luminosity_distance': ('distance', None),
    }

    def get_reparameterisation(self, reparameterisation):
        """Function to get reparameterisations that checks GW defaults and
        aliases
        """
        return get_gw_reparameterisation(reparameterisation)

    def add_default_reparameterisations(self):
        """
        Add default reparameterisations for parameters that have not been
        specified.
        """
        parameters = list(set(self.names)
                          - set(self._reparameterisation.parameters))
        logger.debug(f'Adding default reparameterisations for {parameters}')

        for p in parameters:
            logger.debug(f'Trying to add reparameterisation for {p}')
            if p in self._reparameterisation.parameters:
                logger.debug(f'Parameter {p} is already included')
                continue
            name, extra_params = self.aliases.get(p.lower(), (None, None))
            if name is None:
                logger.debug(f'{p} is not a known GW parameter')
                continue
            if extra_params is not None:
                p = [p] + [ep for ep in extra_params if ep in self.model.names]
            else:
                p = [p]
            prior_bounds = {k: self.model.bounds[k] for k in p}
            reparam, kwargs = get_gw_reparameterisation(name)
            logger.debug(
                f'Add reparameterisation for {p} with config: {kwargs}')
            self._reparameterisation.add_reparameterisation(
                reparam(parameters=p, prior_bounds=prior_bounds, **kwargs))



#!/usr/bin/env python

# Example of using nessai with `reparameterisations` dictionary. This example
# uses the sample model as the half_gaussian example.

import numpy as np
from scipy.stats import norm

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

output = './outdir/reparemterisations_example/'
logger = setup_logger(output=output, log_level='INFO')


class HalfGaussianModel(Model):
    """
    A simple two-dimensional Guassian likelihood with a cut at x=0.
    """
    def __init__(self):
        # Names of parameters to sample
        self.names = ['x', 'y']
        # Prior bounds for each parameter
        self.bounds = {'x': [0, 10], 'y': [-10, 10]}

    def log_prior(self, x):
        """
        Returns log of prior given a live point assuming uniforn
        priors on each parameter.
        """
        log_p = 0.
        # Iterate through each parameter (x and y)
        # since the live points are a structured array we can
        # get each value using just the name
        for n in self.names:
            log_p += (np.log((x[n] >= self.bounds[n][0])
                             & (x[n] <= self.bounds[n][1]))
                      - np.log(self.bounds[n][1] - self.bounds[n][0]))
        return log_p

    def log_likelihood(self, x):
        """
        Returns log likelihood of given live point assuming a Gaussian
        likelihood.
        """
        log_l = 0
        # Use a Guassian logpdf and iterate through the parameters
        for pn in self.names:
            log_l += norm.logpdf(x[pn])
        return log_l


# Configure the normalising flow
flow_config = dict(
        max_epochs=50,
        patience=10,
        model_config=dict(n_blocks=2, n_neurons=4, n_layers=1,
                          kwargs=dict(batch_norm_between_layers=True))
        )

# In this example we use the reparameterisation options to specifiy how each
# parameter should be rescaled.
fp = FlowSampler(HalfGaussianModel(), output=output, flow_config=flow_config,
                 resume=False, expansion_fraction=1.0,
                 reparameterisations={
                     'x': {'reparameterisation': 'inversion',
                           'detect_edges': True},
                     'y': 'default'
                }
)

fp.run()



# -*- coding: utf-8 -*-
"""
Functions related to creating live points and converting to other common
data-types.
"""
import numpy as np
from numpy.lib import recfunctions as rfn


LOGL_DTYPE = 'f8'
DEFAULT_FLOAT_DTYPE = 'f8'


def get_dtype(names, array_dtype=DEFAULT_FLOAT_DTYPE):
    """
    Get a list of tuples containing the dtypes for the structed array

    Parameters
    ----------
    names : list of str
        Names of parameters
    array_dtype : optional
        dtype to use

    Returns
    -------
    list of tuple
        Dtypes as tuples with (field, dtype)
    """
    return [(n, array_dtype) for n in names] \
        + [('logP', array_dtype), ('logL', LOGL_DTYPE)]


def live_points_to_array(live_points, names=None):
    """
    Converts live points to unstructered arrays for training.

    Parameters
    ----------
    live_points : structured_array
        Structured array of live points
    names : list of str or None
        If None all fields in the structed array are added to the dictionary
        else only those included in the list are added.

    Returns
    -------
    np.ndarray
        Unstructed numpy array
    """
    if names is None:
        names = list(live_points.dtype.names)
    return rfn.structured_to_unstructured(live_points[names])


def parameters_to_live_point(parameters, names):
    """
    Take a list or array of parameters for a single live point
    and converts them to a live point.

    Returns an empty array with the correct fields if len(parameters) is zero

    Parameters
    ----------
    parameters : tuple
        Float point values for each parameter
    names : tuple
        Names for each parameter as strings

    Returns
    -------
    structed_array
        Numpy structed array with fields given by names plus logP and logL
    """
    if not len(parameters):
        return np.empty(0, dtype=get_dtype(names, DEFAULT_FLOAT_DTYPE))
    else:
        return np.array((*parameters, 0., 0.),
                        dtype=get_dtype(names, DEFAULT_FLOAT_DTYPE))


def numpy_array_to_live_points(array, names):
    """
    Convert a numpy array to a numpy structure array with the correct fields

    Parameters
    ----------
    array : np.ndarray
        Instance of np.ndarray to converto to a structed array
    names : tuple
        Names for each parameter as strings

    Returns
    -------
    structed_array
        Numpy structed array with fields given by names plus logP and logL
    """
    if array.size == 0:
        return np.empty(0, dtype=get_dtype(names))
    if array.ndim == 1:
        array = array[np.newaxis, :]
    struct_array = np.zeros((array.shape[0]), dtype=get_dtype(names))
    for i, n in enumerate(names):
        struct_array[n] = array[..., i]
    return struct_array


def dict_to_live_points(d):
    """
    Convert a dictionary with parameters names as keys to live points.

    Parameters
    ----------
    d : dict
        Dictionary with parmeters names as keys and values that correspond
        to one or more parameters

    Returns
    -------
    structured_array
        Numpy structed array with fields given by names plus logP and logL
    """
    if isinstance(list(d.values())[0], int):
        N = 1
    else:
        N = len(list(d.values())[0])
    if N == 1:
        return np.array((*list(d.values()), 0., 0.),
                        dtype=get_dtype(d.keys(), DEFAULT_FLOAT_DTYPE))
    else:
        array = np.zeros(N, dtype=get_dtype(list(d.keys())))
        for k, v in d.items():
            array[k] = v
        return array


def live_points_to_dict(live_points, names=None):
    """
    Convert a structured array of live points to a dictionary with
    a key per field.

    Parameters
    ----------
    live_points : structured_array
        Array of live points
    names : list of str or None
        If None all fields in the structed array are added to the dictionary
        else only those included in the list are added.

    Returns
    -------
    dict
        Dictionary of live points
    """
    if names is None:
        names = live_points.dtype.names
    return {f: live_points[f] for f in names}



import os

from nessai.flowsampler import FlowSampler
import torch
import pytest

torch.set_num_threads(1)


def test_sampling_with_rescale(model, flow_config, tmpdir):
    """
    Test sampling with rescaling. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir('w_rescale'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10, max_threads=1)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


def test_sampling_without_rescale(model, flow_config, tmpdir):
    """
    Test sampling without rescaling. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir('wo_rescale'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=False, seed=1234,
                     max_iteration=11, poolsize=10)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


def test_sampling_with_maf(model, flow_config, tmpdir):
    """
    Test sampling with MAF. Checks that flow is trained but does not
    check convergence.
    """
    flow_config['model_config']['ftype'] = 'maf'
    output = str(tmpdir.mkdir('maf'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.parametrize('analytic', [False, True])
def test_sampling_uninformed(model, flow_config, tmpdir, analytic):
    """
    Test running the sampler with the two uninformed proposal methods.
    """
    output = str(tmpdir.mkdir('uninformed'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=None,
                     maximum_uninformed=10, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10,
                     analytic_proposal=analytic)
    fp.run()


def test_sampling_with_n_pool(model, flow_config, tmpdir):
    """
    Test running the sampler with multiprocessing.
    """
    output = str(tmpdir.mkdir('pool'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10, max_threads=3,
                     n_pool=2)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1
    assert os.path.exists(output + '/result.json')


def test_sampling_resume(model, flow_config, tmpdir):
    """
    Test resuming the sampler.
    """
    output = str(tmpdir.mkdir('resume'))
    fp = FlowSampler(model, output=output, resume=True, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10)
    fp.run()
    assert os.path.exists(os.path.join(output, 'nested_sampler_resume.pkl'))

    fp = FlowSampler(model, output=output, resume=True,
                     flow_config=flow_config)
    assert fp.ns.iteration == 11
    fp.ns.max_iteration = 21
    fp.run()
    assert fp.ns.iteration == 21
    assert os.path.exists(
        os.path.join(output, 'nested_sampler_resume.pkl.old'))



# -*- coding: utf-8 -*-
"""
Implementation of Real Non Volume Preserving flows.
"""
import logging

import torch
import torch.nn.functional as F

from nflows.distributions import StandardNormal
from nflows import transforms

from .base import NFlow

logger = logging.getLogger(__name__)


class FlexibleRealNVP(NFlow):
    """
    Implementation of RealNVP.

    This class modifies ``SimpleRealNVP`` from nflows to allows for a custom
    mask to be parsed as a numpy array and allows for MLP to be used
    instead of a ResNet

    > L. Dinh et al., Density estimation using Real NVP, ICLR 2017.

    Parameters
    ----------
    features : int
        Number of features (dimensions) in the data space
    hidden_features : int
        Number of neurons per layer in each neural network
    num_layers : int
        Number of coupling tranformations
    num_blocks_per_layer : int
        Number of layers (or blocks for resnet) per nerual network for
        each coupling transform
    mask : array_like, optional
        Custom mask to use between coupling transforms. Can either be
        a single array with the same length as the number of features or
        and two-dimensional array of shape (# features, # num_layers).
        Must use -1 and 1 to indicate no updated and updated.
    net : {'resnet', 'mlp'}
        Type of neural network to use
    use_volume_preserving : bool, optional (False)
        Use volume preserving flows which use only additiona and no scaling
    activation : function
        Activation function implemented in torch
    dropout_probability : float, optional (0.0)
        Dropout probaiblity used in each layer of the neural network
    batch_norm_within_layers : bool, optional (False)
       Enable or disable batch norm within the neural network for each coupling
       transform
    batch_norm_between_layers : bool, optional (False)
       Enable or disable batch norm between coupling transforms
    linear_transform : {'permutaiton', 'lu', 'svd'}
        Linear transform to use between coupling layers. Not recommended when
        using a custom mask.
    """
    def __init__(
        self,
        features,
        hidden_features,
        num_layers,
        num_blocks_per_layer,
        mask=None,
        net='resnet',
        use_volume_preserving=False,
        activation=F.relu,
        dropout_probability=0.0,
        batch_norm_within_layers=False,
        batch_norm_between_layers=False,
        linear_transform=None,
        distribution=None,
    ):

        if use_volume_preserving:
            coupling_constructor = transforms.AdditiveCouplingTransform
        else:
            coupling_constructor = transforms.AffineCouplingTransform

        if mask is None:
            mask = torch.ones(features)
            mask[::2] = -1
        else:
            if not mask.shape[-1] == features:
                raise RuntimeError('Mask does not match number of features')
            if mask.ndim == 2 and not mask.shape[0] == num_layers:
                raise RuntimeError('Mask does not match number of layers')

            mask = torch.from_numpy(mask.astype('float32'))

        if mask.dim() == 1:
            mask_array = torch.empty([num_layers, features])
            for i in range(num_layers):
                mask_array[i] = mask
                mask *= -1
            mask = mask_array

        def create_linear_transform():
            if linear_transform == 'permutation':
                return transforms.RandomPermutation(features=features)
            elif linear_transform == 'lu':
                return transforms.CompositeTransform([
                    transforms.RandomPermutation(features=features),
                    transforms.LULinear(features, identity_init=True,
                                        using_cache=True)
                ])
            elif linear_transform == 'svd':
                return transforms.CompositeTransform([
                    transforms.RandomPermutation(features=features),
                    transforms.SVDLinear(features, num_householder=10,
                                         identity_init=True)
                ])
            else:
                raise ValueError

        if net.lower() == 'resnet':
            from nflows.nn.nets import ResidualNet

            def create_net(in_features, out_features):
                return ResidualNet(
                    in_features,
                    out_features,
                    hidden_features=hidden_features,
                    num_blocks=num_blocks_per_layer,
                    activation=activation,
                    dropout_probability=dropout_probability,
                    use_batch_norm=batch_norm_within_layers
                    )
        elif net.lower() == 'mlp':
            from .utils import CustomMLP
            if batch_norm_within_layers:
                logger.warning('Batch norm within layers not supported for '
                               'MLP, will be ignored')
            if dropout_probability:
                logger.warning('Dropout not supported for MLP, '
                               'will be ignored')
            hidden_features = num_blocks_per_layer * [hidden_features]

            def create_net(in_features, out_features):
                return CustomMLP(
                        (in_features,),
                        (out_features,),
                        hidden_features,
                        activation=activation)

        else:
            raise RuntimeError(f'Unknown nn type: {net}')

        layers = []
        for i in range(num_layers):
            if linear_transform is not None:
                layers.append(create_linear_transform())
            transform = coupling_constructor(
                mask=mask[i], transform_net_create_fn=create_net
            )
            layers.append(transform)
            if batch_norm_between_layers:
                layers.append(transforms.BatchNorm(features=features))

        if distribution is None:
            distribution = StandardNormal([features])

        super().__init__(
            transform=transforms.CompositeTransform(layers),
            distribution=distribution,
        )
